import json
import os
import sys
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Dict, Generator, List, Optional, Set, Tuple, Union

from ..debug import debug
from ..extension_support import (
    EXT_EXCLUDE_SUFFIXES,
    extension_info,
    package_mfext_all,
    package_mfext_all_descriptions,
)
from ..exception import MetaflowException
from ..metaflow_version import get_version
from ..user_decorators.user_flow_decorator import FlowMutatorMeta
from ..user_decorators.user_step_decorator import UserStepDecoratorMeta
from ..util import get_metaflow_root, walk_without_cycles
from . import ContentType, MFCONTENT_MARKER, MetaflowCodeContentV1Base
from .distribution_support import _ModuleInfo, modules_to_distributions
from .utils import suffix_filter, walk


class MetaflowCodeContentV1(MetaflowCodeContentV1Base):
    METAFLOW_SUFFIXES_LIST = [".py", ".html", ".css", ".js"]

    def __init__(
        self,
        code_dir: str = MetaflowCodeContentV1Base._code_dir,
        other_dir: str = MetaflowCodeContentV1Base._other_dir,
        criteria: Callable[[ModuleType], bool] = lambda x: True,
    ):
        super().__init__(code_dir, other_dir)

        self._metaflow_root = get_metaflow_root()
        self._metaflow_version = get_version()

        self._criteria = criteria

        # We try to find the modules we need to package. We will first look at all modules
        # and apply the criteria to them. Then we will use the most parent module that
        # fits the criteria as the module to package

        # Make a copy since sys.modules could be modified while we load other
        # modules. See https://github.com/Netflix/metaflow/issues/2489
        all_modules = dict(sys.modules)
        modules = filter(lambda x: criteria(x[1]), all_modules.items())
        # Ensure that we see the parent modules first
        modules = sorted(modules, key=lambda x: x[0])
        if modules:
            last_prefix = modules[0][0]
            new_modules = [modules[0]]
            for name, mod in modules[1:]:
                if name.startswith(last_prefix + "."):
                    # This is a submodule of the last module, we can skip it
                    continue
                # Otherwise, we have a new top-level module
                last_prefix = name
                new_modules.append((name, mod))
        else:
            new_modules = []

        self._modules = {}  # type: Dict[str, _ModuleInfo]
        # We do this explicitly module by module to harden it against misbehaving
        # modules like the one in:
        # https://github.com/Netflix/metaflow/issues/2512
        # We will silently ignore modules that are not well built.
        for name, mod in new_modules:
            try:
                minfo = _ModuleInfo(
                    name,
                    set(
                        Path(p).resolve().as_posix()
                        for p in getattr(mod, "__path__", [mod.__file__])
                    ),
                    mod,
                    True,  # This is a Metaflow module (see filter below)
                )
            except:
                continue
            self._modules[name] = minfo

        # Contain metadata information regarding the distributions packaged.
        # This allows Metaflow to "fake" distribution information when packaged
        self._distmetainfo = {}  # type: Dict[str, Dict[str, str]]

        # Maps an absolute path on the filesystem to the path of the file in the
        # archive.
        self._files = {}  # type: Dict[str, str]
        self._files_from_modules = {}  # type: Dict[str, str]

        self._other_files = {}  # type: Dict[str, str]
        self._other_content = {}  # type: Dict[str, bytes]

        debug.package_exec(f"Used system modules found: {str(self._modules)}")

        # Populate with files from the third party modules
        for k, v in self._modules.items():
            self._files_from_modules.update(self._module_files(k, v.root_paths))

        # Figure out the files to package for Metaflow and extensions
        self._cached_metaflow_files = list(self._metaflow_distribution_files())
        self._cached_metaflow_files.extend(list(self._metaflow_extension_files()))

    def create_mfcontent_info(self) -> Dict[str, Any]:
        return {"version": 1, "module_files": list(self._files_from_modules.values())}

    def get_excluded_tl_entries(self) -> List[str]:
        """
        When packaging Metaflow from within an executing Metaflow flow, we need to
        exclude the files that are inserted by this content from being packaged (possibly).

        Use this function to return these files or top-level directories.

        Returns
        -------
        List[str]
            Files or directories to exclude
        """
        return [self._code_dir, self._other_dir]

    def content_names(
        self, content_types: Optional[int] = None
    ) -> Generator[Tuple[str, str], None, None]:
        """
        Detailed list of the content of this MetaflowCodeContent. This will list all files
        (or non files -- for the INFO or CONFIG data for example) present in the archive.

        Parameters
        ----------
        content_types : Optional[int]
            The type of content to get the names of. If None, all content is returned.

        Yields
        ------
        Generator[Tuple[str, str], None, None]
            Path on the filesystem and the name in the archive
        """
        yield from self._content(content_types, generate_value=False)

    def contents(
        self, content_types: Optional[int] = None
    ) -> Generator[Tuple[Union[bytes, str], str], None, None]:
        """
        Very similar to content_names but returns the content of the non-files
        as well as bytes. For files, identical output as content_names

        Parameters
        ----------
        content_types : Optional[int]
            The type of content to get the content of. If None, all content is returned.

        Yields
        ------
        Generator[Tuple[Union[str, bytes], str], None, None]
            Content of the MF content
        """
        yield from self._content(content_types, generate_value=True)

    def show(self) -> str:
        """
        Returns a more human-readable string representation of the content of this
        MetaflowCodeContent. This will not, for example, list all files but summarize what
        is included at a more high level.

        Returns
        -------
        str
            A human-readable string representation of the content of this MetaflowCodeContent
        """
        all_user_step_decorators = {}
        for k, v in UserStepDecoratorMeta.all_decorators().items():
            all_user_step_decorators.setdefault(
                getattr(v, "_original_module", v.__module__), []
            ).append(k)

        all_user_flow_decorators = {}
        for k, v in FlowMutatorMeta.all_decorators().items():
            all_user_flow_decorators.setdefault(
                getattr(v, "_original_module", v.__module__), []
            ).append(k)

        result = []
        if self._metaflow_version:
            result.append(f"\nMetaflow version: {self._metaflow_version}")
        ext_info = extension_info()
        if ext_info["installed"]:
            result.append("\nMetaflow extensions packaged:")
            for ext_name, ext_info in ext_info["installed"].items():
                result.append(
                    f"  - {ext_name} ({ext_info['extension_name']}) @ {ext_info['dist_version']}"
                )

        if self._modules:
            mf_modules = []
            other_modules = []
            for name, info in self._modules.items():
                if info.metaflow_module:
                    mf_modules.append(f"  - {name} @ {', '.join(info.root_paths)}")
                    module_user_step_decorators = [
                        ", ".join(v)
                        for k, v in all_user_step_decorators.items()
                        if k == info.name or k.startswith(info.name + ".")
                    ]
                    module_user_flow_decorators = [
                        ", ".join(v)
                        for k, v in all_user_flow_decorators.items()
                        if k == info.name or k.startswith(info.name + ".")
                    ]
                    if module_user_step_decorators:
                        mf_modules.append(
                            f"    - Provides step decorators: {', '.join(module_user_step_decorators)}"
                        )
                    if module_user_flow_decorators:
                        mf_modules.append(
                            f"    - Provides flow mutators: {', '.join(module_user_flow_decorators)}"
                        )
                else:
                    other_modules.append(f"  - {name} @ {', '.join(info.root_paths)}")
            if mf_modules:
                result.append("\nMetaflow modules:")
                result.extend(mf_modules)
            if other_modules:
                result.append("\nNon-Metaflow packaged modules:")
                result.extend(other_modules)

        return "\n".join(result)

    def add_info(self, info: Dict[str, Any]) -> None:
        """
        Add the content of the INFO file to the Metaflow content

        Parameters
        ----------
        info: Dict[str, Any]
            The content of the INFO file
        """
        info_file_path = os.path.join(self._other_dir, self._info_file)
        if info_file_path in self._other_content:
            raise MetaflowException("INFO file already present in the MF environment")
        self._other_content[info_file_path] = json.dumps(info).encode("utf-8")

    def add_config(self, config: Dict[str, Any]) -> None:
        """
        Add the content of the CONFIG file to the Metaflow content

        Parameters
        ----------
        config: Dict[str, Any]
            The content of the CONFIG file
        """
        config_file_path = os.path.join(self._other_dir, self._config_file)
        if config_file_path in self._other_content:
            raise MetaflowException("CONFIG file already present in the MF environment")
        self._other_content[config_file_path] = json.dumps(config).encode("utf-8")

    def add_module(self, module: ModuleType) -> None:
        """
        Add a python module to the Metaflow content

        Parameters
        ----------
        module_path: ModuleType
            The module to add
        """
        name = module.__name__
        debug.package_exec(f"Adding module {name} to the MF content")
        # If the module is a single file, we handle this here by looking at __file__
        # which will point to the single file. If it is an actual module, __path__
        # will contain the path(s) to the module
        if hasattr(module, "__file__") and module.__file__:
            root_paths = [Path(module.__file__).resolve().as_posix()]
        else:
            root_paths = []
            seen_path_values = set()
            new_paths = module.__spec__.submodule_search_locations
            while new_paths:
                paths = new_paths
                new_paths = []
                for p in paths:
                    if p in seen_path_values:
                        continue
                    if os.path.isdir(p):
                        root_paths.append(Path(p).resolve().as_posix())
                    elif p in sys.path_importer_cache:
                        # We have a path hook that we likely need to call to get the actual path
                        addl_spec = sys.path_importer_cache[p].find_spec(name)
                        if (
                            addl_spec is not None
                            and addl_spec.submodule_search_locations
                        ):
                            new_paths.extend(addl_spec.submodule_search_locations)
                    else:
                        # This may not be as required since it is likely the importer cache has
                        # everything already but just in case, we will also go through the
                        # path hooks and see if we find another one
                        for path_hook in sys.path_hooks:
                            try:
                                finder = path_hook(p)
                                addl_spec = finder.find_spec(name)
                                if (
                                    addl_spec is not None
                                    and addl_spec.submodule_search_locations
                                ):
                                    new_paths.extend(
                                        addl_spec.submodule_search_locations
                                    )
                                    break
                            except ImportError:
                                continue
                    seen_path_values.add(p)
        self._modules[name] = _ModuleInfo(
            name,
            set(root_paths),
            module,
            False,  # This is not a Metaflow module (added by the user manually)
        )
        self._files_from_modules.update(
            self._module_files(name, self._modules[name].root_paths)
        )

    def add_code_file(self, file_path: str, file_name: str) -> None:
        """
        Add a code file to the Metaflow content

        Parameters
        ----------
        file_path: str
            The path to the code file to add (on the filesystem)
        file_name: str
            The path in the archive to add the code file to
        """
        file_path = os.path.realpath(file_path)
        debug.package_exec(
            f"Adding code file {file_path} as {file_name} to the MF content"
        )

        if file_path in self._files and self._files[file_path] != os.path.join(
            self._code_dir, file_name.lstrip("/")
        ):
            raise MetaflowException(
                "File '%s' is already present in the MF content with a different name: '%s'"
                % (file_path, self._files[file_path])
            )
        self._files[file_path] = os.path.join(self._code_dir, file_name.lstrip("/"))

    def add_other_file(self, file_path: str, file_name: str) -> None:
        """
        Add a non-python file to the Metaflow content

        Parameters
        ----------
        file_path: str
            The path to the file to add (on the filesystem)
        file_name: str
            The path in the archive to add the file to
        """
        file_path = os.path.realpath(file_path)
        debug.package_exec(
            f"Adding other file {file_path} as {file_name} to the MF content"
        )
        if file_path in self._other_files and self._other_files[
            file_path
        ] != os.path.join(self._other_dir, file_name.lstrip("/")):
            raise MetaflowException(
                "File %s is already present in the MF content with a different name: %s"
                % (file_path, self._other_files[file_path])
            )
        self._other_files[file_path] = os.path.join(
            self._other_dir, file_name.lstrip("/")
        )

    def _content(
        self, content_types: Optional[int] = None, generate_value: bool = False
    ) -> Generator[Tuple[Union[str, bytes], str], None, None]:
        from ..package import MetaflowPackage  # Prevent circular dependency

        if content_types is None:
            content_types = ContentType.ALL_CONTENT.value

        if content_types & ContentType.CODE_CONTENT.value:
            yield from self._cached_metaflow_files
            yield from self._files.items()
        if content_types & ContentType.MODULE_CONTENT.value:
            yield from self._files_from_modules.items()
        if content_types & ContentType.OTHER_CONTENT.value:
            yield from self._other_files.items()
            if generate_value:
                for k, v in self._other_content.items():
                    yield v, k
                # Include the distribution file too
                yield json.dumps(self._distmetainfo).encode("utf-8"), os.path.join(
                    self._other_dir, self._dist_info_file
                )
                yield json.dumps(self.create_mfcontent_info()).encode(
                    "utf-8"
                ), MFCONTENT_MARKER
            else:
                for k in self._other_content.keys():
                    yield "<generated %s content>" % (os.path.basename(k)), k
                yield "<generated %s content>" % (
                    os.path.basename(self._dist_info_file)
                ), os.path.join(self._other_dir, self._dist_info_file)
                yield "<generated %s content>" % MFCONTENT_MARKER, MFCONTENT_MARKER

    def _metaflow_distribution_files(self) -> Generator[Tuple[str, str], None, None]:
        debug.package_exec("Including Metaflow from '%s'" % self._metaflow_root)
        for path_tuple in walk(
            os.path.join(self._metaflow_root, "metaflow"),
            exclude_hidden=False,
            file_filter=suffix_filter(self.METAFLOW_SUFFIXES_LIST),
        ):
            yield path_tuple[0], os.path.join(self._code_dir, path_tuple[1])

    def _metaflow_extension_files(self) -> Generator[Tuple[str, str], None, None]:
        # Metaflow extensions; for now, we package *all* extensions but this may change
        # at a later date; it is possible to call `package_mfext_package` instead of
        # `package_mfext_all` but in that case, make sure to also add a
        # metaflow_extensions/__init__.py file to properly "close" the metaflow_extensions
        # package and prevent other extensions from being loaded that may be
        # present in the rest of the system
        for path_tuple in package_mfext_all():
            yield path_tuple[0], os.path.join(self._code_dir, path_tuple[1])
        if debug.package:
            ext_info = package_mfext_all_descriptions()
            ext_info = {
                k: {k1: v1 for k1, v1 in v.items() if k1 in ("root_paths",)}
                for k, v in ext_info.items()
            }
            debug.package_exec(f"Metaflow extensions packaged: {ext_info}")

    def _module_files(
        self, name: str, paths: Set[str]
    ) -> Generator[Tuple[str, str], None, None]:
        debug.package_exec(
            "    Looking for distributions for module %s in %s" % (name, paths)
        )
        paths = set(paths)  # Do not modify external paths
        has_init = False
        distributions = modules_to_distributions().get(name)
        prefix_parts = tuple(name.split("."))

        seen_distributions = set()
        if distributions:
            for dist in distributions:
                dist_name = dist.metadata["Name"]  # dist.name not always present
                if dist_name in seen_distributions:
                    continue
                # For some reason, sometimes the same distribution appears twice. We
                # don't need to process twice.
                seen_distributions.add(dist_name)
                debug.package_exec(
                    "    Including distribution '%s' for module '%s'"
                    % (dist_name, name)
                )
                dist_root = str(dist.locate_file(name))
                has_file_in_root = False
                if dist_name not in self._distmetainfo:
                    # Possible that a distribution contributes to multiple modules
                    self._distmetainfo[dist_name] = {
                        # We can add more if needed but these are likely the most
                        # useful (captures, name, version, etc and files which can
                        # be used to find non-python files in the distribution).
                        "METADATA": dist.read_text("METADATA") or "",
                        "RECORD": dist.read_text("RECORD") or "",
                    }
                for file in dist.files or []:
                    # Skip files that do not belong to this module (distribution may
                    # provide multiple modules)
                    if (
                        file.parts[: len(prefix_parts)] != prefix_parts
                        or file.suffix == ".pth"
                        or str(file).startswith("__editable__")
                    ):
                        continue
                    if file.parts[len(prefix_parts)] == "__init__.py":
                        has_init = True
                    has_file_in_root = True
                    # At this point, we know that we are seeing actual files in the
                    # dist_root so we make sure it is as expected
                    if dist_root not in paths:
                        # This is an error because it means that this distribution is
                        # not contributing to the module.
                        raise RuntimeError(
                            "Distribution '%s' is not contributing to module '%s' as "
                            "expected (got '%s' when expected one of %s)"
                            % (dist.metadata["Name"], name, dist_root, paths)
                        )
                    yield str(
                        dist.locate_file(file).resolve().as_posix()
                    ), os.path.join(self._code_dir, *prefix_parts, *file.parts[1:])
                if has_file_in_root:
                    paths.discard(dist_root)

        # Now if there are more paths left in paths, it means there is a non-distribution
        # component to this package which we also include.
        debug.package_exec(
            "    Looking for non-distribution files for module '%s' in %s"
            % (name, paths)
        )
        for path in paths:
            if not Path(path).is_dir():
                # Single file for the module -- this will be something like <name>.py
                yield path, os.path.join(
                    self._code_dir, *prefix_parts[:-1], f"{prefix_parts[-1]}.py"
                )
                has_init = True
            else:
                for root, _, files in walk_without_cycles(path):
                    for file in files:
                        if any(file.endswith(x) for x in EXT_EXCLUDE_SUFFIXES):
                            continue
                        rel_path = os.path.relpath(os.path.join(root, file), path)
                        if rel_path == "__init__.py":
                            has_init = True
                        yield os.path.join(root, file), os.path.join(
                            self._code_dir,
                            name,
                            rel_path,
                        )
        # We now include an empty __init__.py file to close the module and prevent
        # leaks from possible namespace packages
        if not has_init:
            yield os.path.join(
                self._metaflow_root, "metaflow", "extension_support", "_empty_file.py"
            ), os.path.join(self._code_dir, *prefix_parts, "__init__.py")



import os
import sys

from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.propagate import set_global_textmap
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.trace.span import format_trace_id
from opentelemetry.propagate import extract, inject
from functools import wraps
import contextlib
from typing import Dict, List, Optional
from opentelemetry import trace as trace_api, context
from .span_exporter import get_span_exporter

tracer_provider: Optional[TracerProvider] = None


def init_tracing():
    global tracer_provider
    # Disable logging from opentelemetry
    import logging
    logging.getLogger("opentelemetry").setLevel(logging.FATAL)
    if tracer_provider is not None:
        return

    from .propagator import EnvPropagator

    set_global_textmap(EnvPropagator(None))

    span_exporter = get_span_exporter()
    if span_exporter is None:
        return

    tracer_provider = TracerProvider(
        resource=Resource.create(
            {
                SERVICE_NAME: "metaflow",
            }
        )
    )
    trace_api.set_tracer_provider(tracer_provider)

    span_processor = BatchSpanProcessor(span_exporter)
    tracer_provider.add_span_processor(span_processor)

    try:
        from opentelemetry.instrumentation.requests import RequestsInstrumentor

        RequestsInstrumentor().instrument(
            tracer_provider=tracer_provider,
        )
    except ImportError:
        # Its possible that this environment has opentelemetry but not requests
        # instrumentation. We don't want to fail in that case.
        pass


@contextlib.contextmanager
def post_fork():
    global tracer_provider

    tracer_provider = None
    init_tracing()

    token = context.attach(extract(os.environ))
    try:
        tracer = trace_api.get_tracer_provider().get_tracer(__name__)
        with tracer.start_as_current_span(
            "fork", kind=trace_api.SpanKind.SERVER
        ) as span:
            span.set_attribute("cmd", " ".join(sys.argv))
            span.set_attribute("pid", str(os.getpid()))
            yield
    finally:
        context.detach(token)


def cli(name: str):
    def cli_wrap(func):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            global tracer_provider
            init_tracing()

            if tracer_provider is None:
                return func(*args, **kwargs)

            token = context.attach(extract(os.environ))
            try:
                tracer = trace_api.get_tracer_provider().get_tracer(__name__)
                with tracer.start_as_current_span(
                    name, kind=trace_api.SpanKind.SERVER
                ) as span:
                    span.set_attribute("cmd", " ".join(sys.argv))
                    span.set_attribute("pid", str(os.getpid()))
                    return func(*args, **kwargs)
            finally:
                context.detach(token)
                try:
                    tracer_provider.force_flush()
                except Exception as e:  # pylint: disable=broad-except
                    print("TracerProvider failed to flush traces", e, file=sys.stderr)

        return wrapper_func

    return cli_wrap


def inject_tracing_vars(env_dict: Dict[str, str]) -> Dict[str, str]:
    inject(env_dict)
    return env_dict


def get_trace_id() -> str:
    try:
        return format_trace_id(trace_api.get_current_span().get_span_context().trace_id)
    except Exception:
        return ""


@contextlib.contextmanager
def traced(name: str, attrs: Optional[Dict] = None):
    if tracer_provider is None:
        yield
        return
    tracer = trace_api.get_tracer_provider().get_tracer(__name__)
    with tracer.start_as_current_span(name) as span:
        if attrs:
            for k, v in attrs.items():
                span.set_attribute(k, v)
        yield


def tracing(func):
    @wraps(func)
    def wrapper_func(*args, **kwargs):
        if tracer_provider is None:
            return func(*args, **kwargs)

        tracer = trace_api.get_tracer_provider().get_tracer(func.__module__)
        with tracer.start_as_current_span(func.__name__):
            return func(*args, **kwargs)

    return wrapper_func



from __future__ import print_function

import os
import sys

import traceback


# add metaflow module to python path if not already present
myDir = os.path.dirname(os.path.abspath(__file__))
parentDir = os.path.split(os.path.split(myDir)[0])[0]
sys.path.insert(0, parentDir)

from metaflow.sidecar import Message, MessageTypes
from metaflow.plugins import SIDECARS
from metaflow._vendor import click
import metaflow.tracing as tracing


def process_messages(worker_type, worker):
    while True:
        try:
            msg = sys.stdin.readline().strip()
            if msg:
                parsed_msg = Message.deserialize(msg)
                if parsed_msg.msg_type == MessageTypes.INVALID:
                    print(
                        "[sidecar:%s] Invalid message -- skipping: %s"
                        % (worker_type, str(msg))
                    )
                    continue
                else:
                    worker.process_message(parsed_msg)
                    if parsed_msg.msg_type == MessageTypes.SHUTDOWN:
                        break
            else:
                break

        except:  # todo handle other possible exceptions gracefully
            print(
                "[sidecar:%s]: %s" % (worker_type, traceback.format_exc()),
                file=sys.stderr,
            )
            break
    try:
        worker.shutdown()
    except:
        pass


@click.command(help="Initialize workers")
@tracing.cli("sidecar")
@click.argument("worker-type")
def main(worker_type):
    sidecar_type = SIDECARS.get(worker_type)
    if sidecar_type is not None:
        worker_class = sidecar_type.get_worker()
        if worker_class is not None:
            process_messages(worker_type, worker_class())
        else:
            print(
                "[sidecar:%s] Sidecar does not have associated worker" % worker_type,
                file=sys.stderr,
            )
    else:
        print("Unrecognized sidecar_process: %s" % worker_type, file=sys.stderr)


if __name__ == "__main__":
    main()



"""
JSONViewer component for displaying JSON data with syntax highlighting and collapsible sections.
"""

from typing import Any, Optional, Union
from .card import MetaflowCardComponent, with_default_component_id
from .renderer_tools import render_safely
import json
from metaflow._vendor import yaml


class JSONViewer(MetaflowCardComponent):
    """
    A component for displaying JSON data with syntax highlighting and collapsible sections.

    This component provides a rich view of JSON data with proper formatting, syntax highlighting,
    and the ability to collapse/expand sections for better readability.

    Example:
    ```python
    from metaflow.cards import JSONViewer
    from metaflow import current

    data = {
        "user": {"name": "Alice", "age": 30},
        "items": [{"id": 1, "name": "Item 1"}, {"id": 2, "name": "Item 2"}],
        "metadata": {"created": "2024-01-01", "version": "1.0"}
    }

    json_viewer = JSONViewer(data, collapsible=True, max_height="400px")
    current.card.append(json_viewer)
    ```

    Parameters
    ----------
    data : Any
        The data to display as JSON. Will be serialized using json.dumps().
    collapsible : bool, default True
        Whether to make the JSON viewer collapsible.
    max_height : str, optional
        Maximum height for the viewer (CSS value like "300px" or "20rem").
    show_copy_button : bool, default True
        Whether to show a copy-to-clipboard button.
    """

    type = "jsonViewer"

    REALTIME_UPDATABLE = True

    def __init__(
        self,
        data: Any,
        collapsible: bool = True,
        max_height: Optional[str] = None,
        show_copy_button: bool = True,
        title: Optional[str] = None,
    ):
        self._data = data
        self._collapsible = collapsible
        self._max_height = max_height
        self._show_copy_button = show_copy_button
        self._title = title

    def update(self, data: Any):
        """
        Update the JSON data.

        Parameters
        ----------
        data : Any
            New data to display as JSON.
        """
        self._data = data

    @with_default_component_id
    @render_safely
    def render(self):
        # Serialize data to JSON string
        try:
            if isinstance(self._data, str):
                # If already a string, try to parse and re-serialize for formatting
                try:
                    parsed = json.loads(self._data)
                    json_string = json.dumps(parsed, indent=2, ensure_ascii=False)
                except json.JSONDecodeError:
                    # If not valid JSON, treat as plain string
                    json_string = json.dumps(self._data, indent=2, ensure_ascii=False)
            else:
                json_string = json.dumps(
                    self._data, indent=2, ensure_ascii=False, default=str
                )
        except Exception as e:
            # Fallback for non-serializable objects
            json_string = json.dumps(
                {"error": f"Could not serialize data: {str(e)}"}, indent=2
            )

        data = {
            "type": self.type,
            "id": self.component_id,
            "json_string": json_string,
            "collapsible": self._collapsible,
            "show_copy_button": self._show_copy_button,
            "title": self._title or "JSON",
        }

        if self._max_height:
            data["max_height"] = self._max_height

        return data


class YAMLViewer(MetaflowCardComponent):
    """
    A component for displaying YAML data with syntax highlighting and collapsible sections.

    This component provides a rich view of YAML data with proper formatting and syntax highlighting.

    Example:
    ```python
    from metaflow.cards import YAMLViewer
    from metaflow import current

    data = {
        "database": {
            "host": "localhost",
            "port": 5432,
            "credentials": {"username": "admin", "password": "secret"}
        },
        "features": ["auth", "logging", "monitoring"]
    }

    yaml_viewer = YAMLViewer(data, collapsible=True)
    current.card.append(yaml_viewer)
    ```

    Parameters
    ----------
    data : Any
        The data to display as YAML. Will be serialized to YAML format.
    collapsible : bool, default True
        Whether to make the YAML viewer collapsible.
    max_height : str, optional
        Maximum height for the viewer (CSS value like "300px" or "20rem").
    show_copy_button : bool, default True
        Whether to show a copy-to-clipboard button.
    """

    type = "yamlViewer"

    REALTIME_UPDATABLE = True

    def __init__(
        self,
        data: Any,
        collapsible: bool = True,
        max_height: Optional[str] = None,
        show_copy_button: bool = True,
        title: Optional[str] = None,
    ):
        self._data = data
        self._collapsible = collapsible
        self._max_height = max_height
        self._show_copy_button = show_copy_button
        self._title = title

    def update(self, data: Any):
        """
        Update the YAML data.

        Parameters
        ----------
        data : Any
            New data to display as YAML.
        """
        self._data = data

    def _to_yaml_string(self, data: Any) -> str:
        """
        Convert data to YAML string format using vendored YAML module.
        """
        try:
            if isinstance(data, str):
                # Try to parse as JSON first, then convert to YAML
                try:
                    import json

                    parsed = json.loads(data)
                    yaml_result = yaml.dump(
                        parsed, default_flow_style=False, indent=2, sort_keys=False
                    )
                    return (
                        str(yaml_result)
                        if yaml_result is not None
                        else "# Empty YAML result"
                    )
                except json.JSONDecodeError:
                    # If not JSON, return as-is
                    return data
            else:
                yaml_result = yaml.dump(
                    data, default_flow_style=False, indent=2, sort_keys=False
                )
                return (
                    str(yaml_result)
                    if yaml_result is not None
                    else "# Empty YAML result"
                )
        except Exception as e:
            # Fallback to JSON on any error
            import json

            return f"# Error converting to YAML: {str(e)}\n{json.dumps(data, indent=2, default=str)}"

    @with_default_component_id
    @render_safely
    def render(self):
        yaml_string = self._to_yaml_string(self._data)

        data = {
            "type": self.type,
            "id": self.component_id,
            "yaml_string": yaml_string,
            "collapsible": self._collapsible,
            "show_copy_button": self._show_copy_button,
            "title": self._title or "YAML",
        }

        if self._max_height:
            data["max_height"] = self._max_height

        return data



from metaflow.exception import MetaflowException


class AirflowException(MetaflowException):
    headline = "Airflow Exception"

    def __init__(self, msg):
        super().__init__(msg)


class NotSupportedException(MetaflowException):
    headline = "Not yet supported with Airflow"



import abc
from typing import Dict


class SecretsProvider(abc.ABC):
    TYPE = None

    @abc.abstractmethod
    def get_secret_as_dict(self, secret_id, options={}, role=None) -> Dict[str, str]:
        """Retrieve the secret from secrets backend, and return a dictionary of
        environment variables."""


from .secrets_func import get_secret



# this file can be overridden by extensions as is (e.g. metaflow-nflx-extensions)
from metaflow.exception import MetaflowException


class ParserValueError(MetaflowException):
    headline = "Value error"


def requirements_txt_parser(content: str):
    """
    Parse non-comment lines from a requirements.txt file as strictly valid
    PEP 508 requirements.

    Recognizes direct references (e.g. "my_lib @ git+https://..."), extras
    (e.g. "requests[security]"), and version specifiers (e.g. "==2.0"). If
    the package name is "python", its specifier is stored in the "python"
    key instead of "packages".

    Parameters
    ----------
    content : str
        Contents of a requirements.txt file.

    Returns
    -------
    dict
        A dictionary with two keys:
            - "packages": dict(str -> str)
              Mapping from package name (plus optional extras/references) to a
              version specifier string.
            - "python": str or None
              The Python version constraints if present, otherwise None.

    Raises
    ------
     ParserValueError
        If a requirement line is invalid PEP 508 or if environment markers are
        detected, or if multiple Python constraints are specified.
    """
    import re
    from metaflow._vendor.packaging.requirements import Requirement, InvalidRequirement

    parsed = {"packages": {}, "python": None}

    inline_comment_pattern = re.compile(r"\s+#.*$")
    for line in content.splitlines():
        line = line.strip()

        # support Rye lockfiles by skipping lines not compliant with requirements
        if line == "-e file:.":
            continue

        if not line or line.startswith("#"):
            continue

        line = inline_comment_pattern.sub("", line).strip()
        if not line:
            continue

        try:
            req = Requirement(line)
        except InvalidRequirement:
            raise ParserValueError(f"Not a valid PEP 508 requirement: '{line}'")

        if req.marker is not None:
            raise ParserValueError(
                "Environment markers (e.g. 'platform_system==\"Linux\"') "
                f"are not supported for line: '{line}'"
            )

        dep_key = req.name
        if req.extras:
            dep_key += f"[{','.join(req.extras)}]"
        if req.url:
            dep_key += f"@{req.url}"

        dep_spec = str(req.specifier).lstrip(" =")

        if req.name.lower() == "python":
            if parsed["python"] is not None and dep_spec:
                raise ParserValueError(
                    f"Multiple Python version specs not allowed: '{line}'"
                )
            parsed["python"] = dep_spec or None
        else:
            parsed["packages"][dep_key] = dep_spec

    return parsed


def pyproject_toml_parser(content: str):
    """
    Parse a pyproject.toml file per PEP 621.

    Reads the 'requires-python' and 'dependencies' fields from the "[project]" section.
    Each dependency line must be a valid PEP 508 requirement. If the package name is
    "python", its specifier is stored in the "python" key instead of "packages".

    Parameters
    ----------
    content : str
        Contents of a pyproject.toml file.

    Returns
    -------
    dict
        A dictionary with two keys:
            - "packages": dict(str -> str)
              Mapping from package name (plus optional extras/references) to a
              version specifier string.
            - "python": str or None
              The Python version constraints if present, otherwise None.

    Raises
    ------
    RuntimeError
        If no TOML library (tomllib in Python 3.11+ or tomli in earlier versions) is found.
     ParserValueError
        If a dependency is not valid PEP 508, if environment markers are used, or if
        multiple Python constraints are specified.
    """
    try:
        import tomllib as toml  # Python 3.11+
    except ImportError:
        try:
            import tomli as toml  # Python < 3.11 (requires "tomli" package)
        except ImportError:
            raise RuntimeError(
                "Could not import a TOML library. For Python <3.11, please install 'tomli'."
            )
    from metaflow._vendor.packaging.requirements import Requirement, InvalidRequirement

    data = toml.loads(content)

    project = data.get("project", {})
    requirements = project.get("dependencies", [])
    requires_python = project.get("requires-python")

    parsed = {"packages": {}, "python": None}

    if requires_python is not None:
        # If present, store verbatim; note that PEP 621 does not necessarily
        # require "python" to be a dependency in the usual sense.
        # Example: "requires-python" = ">=3.7,<4"
        parsed["python"] = requires_python.lstrip("=").strip()

    for dep_line in requirements:
        dep_line_stripped = dep_line.strip()
        try:
            req = Requirement(dep_line_stripped)
        except InvalidRequirement:
            raise ParserValueError(
                f"Not a valid PEP 508 requirement: '{dep_line_stripped}'"
            )

        if req.marker is not None:
            raise ParserValueError(
                f"Environment markers not supported for line: '{dep_line_stripped}'"
            )

        dep_key = req.name
        if req.extras:
            dep_key += f"[{','.join(req.extras)}]"
        if req.url:
            dep_key += f"@{req.url}"

        dep_spec = str(req.specifier).lstrip("=")

        if req.name.lower() == "python":
            if parsed["python"] is not None and dep_spec:
                raise ParserValueError(
                    f"Multiple Python version specs not allowed: '{dep_line_stripped}'"
                )
            parsed["python"] = dep_spec or None
        else:
            parsed["packages"][dep_key] = dep_spec

    return parsed


def conda_environment_yml_parser(content: str):
    """
    Parse a minimal environment.yml file under strict assumptions.

    The file must contain a 'dependencies:' line, after which each dependency line
    appears with a '- ' prefix. Python can appear as 'python=3.9', etc.; other
    packages as 'numpy=1.21.2' or simply 'numpy'. Non-compliant lines raise  ParserValueError.

    Parameters
    ----------
    content : str
        Contents of a environment.yml file.

    Returns
    -------
    dict
        A dictionary with keys:
        {
            "packages": dict(str -> str),
            "python": str or None
        }

    Raises
    ------
     ParserValueError
        If the file has malformed lines or unsupported sections.
    """
    import re

    packages = {}
    python_version = None

    inside_dependencies = False

    # Basic pattern for lines like "numpy=1.21.2"
    # Group 1: package name
    # Group 2: optional operator + version (could be "=1.21.2", "==1.21.2", etc.)
    line_regex = re.compile(r"^([A-Za-z0-9_\-\.]+)(\s*[=<>!~].+\s*)?$")
    inline_comment_pattern = re.compile(r"\s+#.*$")

    for line in content.splitlines():
        line = line.strip()
        if not line or line.startswith("#"):
            continue

        line = inline_comment_pattern.sub("", line).strip()
        if not line:
            continue

        if line.lower().startswith("dependencies:"):
            inside_dependencies = True
            continue

        if inside_dependencies and not line.startswith("-"):
            inside_dependencies = False
            continue

        if not inside_dependencies:
            continue

        dep_line = line.lstrip("-").strip()
        if dep_line.endswith(":"):
            raise ParserValueError(
                f"Unsupported subsection '{dep_line}' in environment.yml."
            )

        match = line_regex.match(dep_line)
        if not match:
            raise ParserValueError(
                f"Line '{dep_line}' is not a valid conda package specifier."
            )

        pkg_name, pkg_version_part = match.groups()
        version_spec = pkg_version_part.strip() if pkg_version_part else ""

        if version_spec.startswith("="):
            version_spec = version_spec.lstrip("=").strip()

        if pkg_name.lower() == "python":
            if python_version is not None and version_spec:
                raise ParserValueError(
                    f"Multiple Python version specs detected: '{dep_line}'"
                )
            python_version = version_spec
        else:
            packages[pkg_name] = version_spec

    return {"packages": packages, "python": python_version}



import sys
import json
from typing import ClassVar, Optional, List

from metaflow.plugins.aws.step_functions.step_functions import StepFunctions
from metaflow.runner.deployer import DeployedFlow, TriggeredRun

from metaflow.runner.utils import get_lower_level_group, handle_timeout, temporary_fifo


class StepFunctionsTriggeredRun(TriggeredRun):
    """
    A class representing a triggered AWS Step Functions state machine execution.
    """

    def terminate(self, **kwargs) -> bool:
        """
        Terminate the running state machine execution.

        Parameters
        ----------
        authorize : str, optional, default None
            Authorize the termination with a production token.

        Returns
        -------
        bool
            True if the command was successful, False otherwise.
        """
        _, run_id = self.pathspec.split("/")

        # every subclass needs to have `self.deployer_kwargs`
        command = get_lower_level_group(
            self.deployer.api,
            self.deployer.top_level_kwargs,
            self.deployer.TYPE,
            self.deployer.deployer_kwargs,
        ).terminate(run_id=run_id, **kwargs)

        pid = self.deployer.spm.run_command(
            [sys.executable, *command],
            env=self.deployer.env_vars,
            cwd=self.deployer.cwd,
            show_output=self.deployer.show_output,
        )

        command_obj = self.deployer.spm.get(pid)
        command_obj.sync_wait()
        return command_obj.process.returncode == 0


class StepFunctionsDeployedFlow(DeployedFlow):
    """
    A class representing a deployed AWS Step Functions state machine.
    """

    TYPE: ClassVar[Optional[str]] = "step-functions"

    @classmethod
    def list_deployed_flows(cls, flow_name: Optional[str] = None):
        """
        This method is not currently implemented for Step Functions.

        Raises
        ------
        NotImplementedError
            This method is not implemented for Step Functions.
        """
        raise NotImplementedError(
            "list_deployed_flows is not implemented for StepFunctions"
        )

    @classmethod
    def from_deployment(cls, identifier: str, metadata: Optional[str] = None):
        """
        This method is not currently implemented for Step Functions.

        Raises
        ------
        NotImplementedError
            This method is not implemented for Step Functions.
        """
        raise NotImplementedError(
            "from_deployment is not implemented for StepFunctions"
        )

    @classmethod
    def get_triggered_run(
        cls, identifier: str, run_id: str, metadata: Optional[str] = None
    ):
        """
        This method is not currently implemented for Step Functions.

        Raises
        ------
        NotImplementedError
            This method is not implemented for Step Functions.
        """
        raise NotImplementedError(
            "get_triggered_run is not implemented for StepFunctions"
        )

    @property
    def production_token(self: DeployedFlow) -> Optional[str]:
        """
        Get the production token for the deployed flow.

        Returns
        -------
        str, optional
            The production token, None if it cannot be retrieved.
        """
        try:
            _, production_token = StepFunctions.get_existing_deployment(
                self.deployer.name
            )
            return production_token
        except TypeError:
            return None

    def list_runs(
        self, states: Optional[List[str]] = None
    ) -> List[StepFunctionsTriggeredRun]:
        """
        List runs of the deployed flow.

        Parameters
        ----------
        states : List[str], optional, default None
            A list of states to filter the runs by. Allowed values are:
            RUNNING, SUCCEEDED, FAILED, TIMED_OUT, ABORTED.
            If not provided, all states will be considered.

        Returns
        -------
        List[StepFunctionsTriggeredRun]
            A list of TriggeredRun objects representing the runs of the deployed flow.

        Raises
        ------
        ValueError
            If any of the provided states are invalid or if there are duplicate states.
        """
        VALID_STATES = {"RUNNING", "SUCCEEDED", "FAILED", "TIMED_OUT", "ABORTED"}

        if states is None:
            states = []

        unique_states = set(states)
        if not unique_states.issubset(VALID_STATES):
            invalid_states = unique_states - VALID_STATES
            raise ValueError(
                f"Invalid states found: {invalid_states}. Valid states are: {VALID_STATES}"
            )

        if len(states) != len(unique_states):
            raise ValueError("Duplicate states are not allowed")

        triggered_runs = []
        executions = StepFunctions.list(self.deployer.name, states)

        for e in executions:
            run_id = "sfn-%s" % e["name"]
            tr = StepFunctionsTriggeredRun(
                deployer=self.deployer,
                content=json.dumps(
                    {
                        "metadata": self.deployer.metadata,
                        "pathspec": "/".join((self.deployer.flow_name, run_id)),
                        "name": run_id,
                    }
                ),
            )
            triggered_runs.append(tr)

        return triggered_runs

    def delete(self, **kwargs) -> bool:
        """
        Delete the deployed state machine.

        Parameters
        ----------
        authorize : str, optional, default None
            Authorize the deletion with a production token.

        Returns
        -------
        bool
            True if the command was successful, False otherwise.
        """
        command = get_lower_level_group(
            self.deployer.api,
            self.deployer.top_level_kwargs,
            self.deployer.TYPE,
            self.deployer.deployer_kwargs,
        ).delete(**kwargs)

        pid = self.deployer.spm.run_command(
            [sys.executable, *command],
            env=self.deployer.env_vars,
            cwd=self.deployer.cwd,
            show_output=self.deployer.show_output,
        )

        command_obj = self.deployer.spm.get(pid)
        command_obj.sync_wait()
        return command_obj.process.returncode == 0

    def trigger(self, **kwargs) -> StepFunctionsTriggeredRun:
        """
        Trigger a new run for the deployed flow.

        Parameters
        ----------
        **kwargs : Any
            Additional arguments to pass to the trigger command,
            `Parameters` in particular

        Returns
        -------
        StepFunctionsTriggeredRun
            The triggered run instance.

        Raises
        ------
        Exception
            If there is an error during the trigger process.
        """
        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            # every subclass needs to have `self.deployer_kwargs`
            command = get_lower_level_group(
                self.deployer.api,
                self.deployer.top_level_kwargs,
                self.deployer.TYPE,
                self.deployer.deployer_kwargs,
            ).trigger(deployer_attribute_file=attribute_file_path, **kwargs)

            pid = self.deployer.spm.run_command(
                [sys.executable, *command],
                env=self.deployer.env_vars,
                cwd=self.deployer.cwd,
                show_output=self.deployer.show_output,
            )

            command_obj = self.deployer.spm.get(pid)
            content = handle_timeout(
                attribute_file_fd, command_obj, self.deployer.file_read_timeout
            )

            command_obj.sync_wait()
            if command_obj.process.returncode == 0:
                return StepFunctionsTriggeredRun(
                    deployer=self.deployer, content=content
                )

        raise Exception(
            "Error triggering %s on %s for %s"
            % (
                self.deployer.name,
                self.deployer.TYPE,
                self.deployer.flow_file,
            )
        )



import json


# Define message enums
# Unfortunately we can't use enums because they are not supported
# officially in Python2
# INVALID: Not a valid message
# MUST_SEND: Will attempt to send until successful and not send any BEST_EFFORT
#            messages until then. A newer MUST_SEND message will take precedence on
#            any currently unsent one
# BEST_EFFORT: Will try to send once and drop if not possible
# SHUTDOWN: Signal termination; also best effort
class MessageTypes(object):
    INVALID, MUST_SEND, BEST_EFFORT, SHUTDOWN = range(1, 5)


class Message(object):
    def __init__(self, msg_type, payload):
        self.msg_type = msg_type
        self.payload = payload

    def serialize(self):
        msg = {
            "msg_type": self.msg_type,
            "payload": self.payload,
        }
        return json.dumps(msg) + "\n"

    @staticmethod
    def deserialize(json_msg):
        try:
            return Message(**json.loads(json_msg))
        except json.decoder.JSONDecodeError:
            return Message(MessageTypes.INVALID, None)



# This class provides a global singleton `cli_args` which stores the `top` and
# `step` level options for the metaflow CLI. This allows decorators to have
# access to the CLI options instead of relying (solely) on the click context.
# TODO: We have two CLIArgs:
#  - this one, which captures the top level and step-level options passed to the
#    step command and is used primarily for UBF to replicate the exact command
#    line passed
#  - one in runtime.py which is used to construct the step command and modified by
#    runtime_step_cli. Both are similar in nature and should be unified in some way
#
# TODO: dict_to_cli_options uses shlex which causes some issues with this as
# well as the converting of options in runtime.py. We should make it so that we
# can properly shlex things and un-shlex when using. Ideally this should all be
# done in one place.
#
# NOTE: There is an important between these two as well:
#  - this one will include local_config_file whereas the other one WILL NOT.
#    This is because this is used when constructing the parallel UBF command which
#    executes locally and therefore needs the local_config_file but the other (remote)
#    commands do not.

from .user_configs.config_options import ConfigInput
from .util import to_unicode


class CLIArgs(object):
    def __init__(self):
        self._top_kwargs = {}
        self._step_kwargs = {}

    def _set_step_kwargs(self, kwargs):
        self._step_kwargs = kwargs

    def _set_top_kwargs(self, kwargs):
        self._top_kwargs = kwargs

    @property
    def top_kwargs(self):
        return self._top_kwargs

    @property
    def step_kwargs(self):
        return self._step_kwargs

    def step_command(
        self, executable, script, step_name, top_kwargs=None, step_kwargs=None
    ):
        cmd = [executable, "-u", script]
        if top_kwargs is None:
            top_kwargs = self._top_kwargs
        if step_kwargs is None:
            step_kwargs = self._step_kwargs

        top_args_list = list(self._options(top_kwargs))
        cmd.extend(top_args_list)
        cmd.extend(["step", step_name])
        step_args_list = list(self._options(step_kwargs))
        cmd.extend(step_args_list)

        return cmd

    @staticmethod
    def _options(mapping):
        for k, v in mapping.items():

            # None or False arguments are ignored
            # v needs to be explicitly False, not falsy, e.g. 0 is an acceptable value
            if v is None or v is False:
                continue

            # we need special handling for 'with' since it is a reserved
            # keyword in Python, so we call it 'decospecs' in click args
            if k == "decospecs":
                k = "with"
            if k in ("config", "config_value"):
                # Special handling here since we gather them all in one option but actually
                # need to send them one at a time using --config-value <name> kv.<name>.
                # Note it can be either config or config_value depending
                # on click processing order.
                for config_name in v.keys():
                    yield "--config-value"
                    yield to_unicode(config_name)
                    yield to_unicode(ConfigInput.make_key_name(config_name))
                continue
            k = k.replace("_", "-")
            v = v if isinstance(v, (list, tuple, set)) else [v]
            for value in v:
                yield "--%s" % k
                if not isinstance(value, bool):
                    yield to_unicode(value)


cli_args = CLIArgs()



import json
from metaflow.metadata_provider import MetaDatum


def _save_metadata(
    metadata_provider,
    run_id,
    step_name,
    task_id,
    attempt_id,
    card_uuid,
    save_metadata,
):
    entries = [
        MetaDatum(
            field=card_uuid,
            value=json.dumps(save_metadata),
            type="card-info",
            tags=["attempt_id:{0}".format(attempt_id)],
        )
    ]
    metadata_provider.register_metadata(run_id, step_name, task_id, entries)



import bz2
import concurrent.futures
import io
import json
import os
import shutil
import subprocess
import sys
import tarfile
import time
import platform
from urllib.error import URLError
from urllib.request import urlopen
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR, CONDA_USE_FAST_INIT
from metaflow.packaging_sys import MetaflowCodeContent, ContentType
from metaflow.plugins import DATASTORES
from metaflow.plugins.pypi.utils import MICROMAMBA_MIRROR_URL, MICROMAMBA_URL
from metaflow.util import which
from urllib.request import Request
import warnings

from . import MAGIC_FILE, _datastore_packageroot

FAST_INIT_BIN_URL = "https://fast-flow-init.outerbounds.sh/{platform}/latest"

# Bootstraps a valid conda virtual environment composed of conda and pypi packages


def timer(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        duration = time.time() - start_time
        # print(f"Time taken for {func.__name__}: {duration:.2f} seconds")
        return result

    return wrapper


if __name__ == "__main__":

    def run_cmd(cmd, stdin_str=None):
        result = subprocess.run(
            cmd,
            shell=True,
            input=stdin_str,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        if result.returncode != 0:
            print(f"Bootstrap failed while executing: {cmd}")
            print("Stdout:", result.stdout)
            print("Stderr:", result.stderr)
            sys.exit(1)

    @timer
    def install_fast_initializer(architecture):
        import gzip

        fast_initializer_path = os.path.join(
            os.getcwd(), "fast-initializer", "bin", "fast-initializer"
        )

        if which("fast-initializer"):
            return which("fast-initializer")
        if os.path.exists(fast_initializer_path):
            os.environ["PATH"] += os.pathsep + os.path.dirname(fast_initializer_path)
            return fast_initializer_path

        url = FAST_INIT_BIN_URL.format(platform=architecture)

        # Prepare directory once
        os.makedirs(os.path.dirname(fast_initializer_path), exist_ok=True)

        # Download and decompress in one go
        def _download_and_extract(url):
            headers = {
                "Accept-Encoding": "gzip, deflate, br",
                "Connection": "keep-alive",
                "User-Agent": "python-urllib",
            }

            max_retries = 3
            for attempt in range(max_retries):
                try:
                    req = Request(url, headers=headers)
                    with urlopen(req) as response:
                        with gzip.GzipFile(fileobj=response) as gz:
                            with open(fast_initializer_path, "wb") as f:
                                f.write(gz.read())
                    break
                except (URLError, IOError) as e:
                    if attempt == max_retries - 1:
                        raise Exception(
                            f"Failed to download fast-initializer after {max_retries} attempts: {e}"
                        )
                    time.sleep(2**attempt)

        _download_and_extract(url)

        # Set executable permission
        os.chmod(fast_initializer_path, 0o755)

        # Update PATH only once at the end
        os.environ["PATH"] += os.pathsep + os.path.dirname(fast_initializer_path)
        return fast_initializer_path

    @timer
    def install_micromamba(architecture):
        micromamba_dir = os.path.join(os.getcwd(), "micromamba")
        micromamba_path = os.path.join(micromamba_dir, "bin", "micromamba")

        if which("micromamba"):
            return which("micromamba")
        if os.path.exists(micromamba_path):
            os.environ["PATH"] += os.pathsep + os.path.dirname(micromamba_path)
            return micromamba_path

        # Download and extract in one go
        url = MICROMAMBA_URL.format(platform=architecture, version="2.0.4")
        mirror_url = MICROMAMBA_MIRROR_URL.format(
            platform=architecture, version="2.0.4"
        )

        # Prepare directory once
        os.makedirs(os.path.dirname(micromamba_path), exist_ok=True)

        # Download and decompress in one go
        def _download_and_extract(url):
            headers = {
                "Accept-Encoding": "gzip, deflate, br",
                "Connection": "keep-alive",
                "User-Agent": "python-urllib",
            }

            max_retries = 3
            for attempt in range(max_retries):
                try:
                    req = Request(url, headers=headers)

                    with urlopen(req) as response:
                        decompressor = bz2.BZ2Decompressor()
                        with warnings.catch_warnings():
                            warnings.filterwarnings(
                                "ignore", category=DeprecationWarning
                            )
                            with tarfile.open(
                                fileobj=io.BytesIO(
                                    decompressor.decompress(response.read())
                                ),
                                mode="r:",
                            ) as tar:
                                member = tar.getmember("bin/micromamba")
                                tar.extract(member, micromamba_dir)
                    break
                except (URLError, IOError) as e:
                    if attempt == max_retries - 1:
                        raise Exception(
                            f"Failed to download micromamba after {max_retries} attempts: {e}"
                        )
                    time.sleep(2**attempt)

        try:
            # first try from mirror
            _download_and_extract(mirror_url)
        except Exception:
            # download from mirror failed, try official source before failing.
            _download_and_extract(url)

        # Set executable permission
        os.chmod(micromamba_path, 0o755)

        # Update PATH only once at the end
        os.environ["PATH"] += os.pathsep + os.path.dirname(micromamba_path)
        return micromamba_path

    @timer
    def download_conda_packages(storage, packages, dest_dir):
        def process_conda_package(args):
            # Ensure that conda packages go into architecture specific folders.
            # The path looks like REPO/CHANNEL/CONDA_SUBDIR/PACKAGE. We trick
            # Micromamba into believing that all packages are coming from a local
            # channel - the only hurdle is ensuring that packages are organised
            # properly.
            key, tmpfile, dest_dir = args
            dest = os.path.join(dest_dir, "/".join(key.split("/")[-2:]))
            os.makedirs(os.path.dirname(dest), exist_ok=True)
            shutil.move(tmpfile, dest)

        os.makedirs(dest_dir, exist_ok=True)
        with storage.load_bytes([package["path"] for package in packages]) as results:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                executor.map(
                    process_conda_package,
                    [(key, tmpfile, dest_dir) for key, tmpfile, _ in results],
                )
            # for key, tmpfile, _ in results:

            #     # TODO: consider RAM disk
            #     dest = os.path.join(dest_dir, "/".join(key.split("/")[-2:]))
            #     os.makedirs(os.path.dirname(dest), exist_ok=True)
            #     shutil.move(tmpfile, dest)
        return dest_dir

    @timer
    def download_pypi_packages(storage, packages, dest_dir):
        def process_pypi_package(args):
            key, tmpfile, dest_dir = args
            dest = os.path.join(dest_dir, os.path.basename(key))
            shutil.move(tmpfile, dest)

        os.makedirs(dest_dir, exist_ok=True)
        with storage.load_bytes([package["path"] for package in packages]) as results:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                executor.map(
                    process_pypi_package,
                    [(key, tmpfile, dest_dir) for key, tmpfile, _ in results],
                )
            # for key, tmpfile, _ in results:
            #     dest = os.path.join(dest_dir, os.path.basename(key))
            #     shutil.move(tmpfile, dest)
        return dest_dir

    @timer
    def create_conda_environment(prefix, conda_pkgs_dir):
        cmd = f'''set -e;
            tmpfile=$(mktemp);
            echo "@EXPLICIT" > "$tmpfile";
            ls -d {conda_pkgs_dir}/*/* >> "$tmpfile";
            export PATH=$PATH:$(pwd)/micromamba;
            export CONDA_PKGS_DIRS=$(pwd)/micromamba/pkgs;
            export MAMBA_NO_LOW_SPEED_LIMIT=1;
            export MAMBA_USE_INDEX_CACHE=1;
            export MAMBA_NO_PROGRESS_BARS=1;
            export CONDA_FETCH_THREADS=1;
            micromamba create --yes --offline --no-deps \
                --safety-checks=disabled --no-extra-safety-checks \
                --prefix {prefix} --file "$tmpfile" \
                --no-pyc --no-rc --always-copy;
            rm "$tmpfile"'''
        run_cmd(cmd)

    @timer
    def install_pypi_packages(prefix, pypi_pkgs_dir):
        cmd = f"""set -e;
            export PATH=$PATH:$(pwd)/micromamba;
            export CONDA_PKGS_DIRS=$(pwd)/micromamba/pkgs;
            micromamba run --prefix {prefix} python -m pip --disable-pip-version-check \
                install --root-user-action=ignore --no-compile --no-index \
                --no-cache-dir --no-deps --prefer-binary \
                --find-links={pypi_pkgs_dir}  --no-user \
                --no-warn-script-location --no-input \
                {pypi_pkgs_dir}/*.whl
            """
        run_cmd(cmd)

    @timer
    def setup_environment(
        architecture, storage, env, prefix, conda_pkgs_dir, pypi_pkgs_dir
    ):
        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            # install micromamba, download conda and pypi packages in parallel
            futures = {
                "micromamba": executor.submit(install_micromamba, architecture),
                "conda_pkgs": executor.submit(
                    download_conda_packages, storage, env["conda"], conda_pkgs_dir
                ),
            }
            if "pypi" in env:
                futures["pypi_pkgs"] = executor.submit(
                    download_pypi_packages, storage, env["pypi"], pypi_pkgs_dir
                )

            # create conda environment after micromamba is installed and conda packages are downloaded
            done, _ = concurrent.futures.wait(
                [futures["micromamba"], futures["conda_pkgs"]],
                return_when=concurrent.futures.ALL_COMPLETED,
            )

            for future in done:
                future.result()

            # start conda environment creation
            futures["conda_env"] = executor.submit(
                create_conda_environment, prefix, conda_pkgs_dir
            )

            if "pypi" in env:
                # install pypi packages after conda environment is created and pypi packages are downloaded
                done, _ = concurrent.futures.wait(
                    [futures["conda_env"], futures["pypi_pkgs"]],
                    return_when=concurrent.futures.ALL_COMPLETED,
                )

                for future in done:
                    future.result()

                # install pypi packages
                futures["pypi_install"] = executor.submit(
                    install_pypi_packages, prefix, pypi_pkgs_dir
                )
                # wait for pypi packages to be installed
                futures["pypi_install"].result()
            else:
                # wait for conda environment to be created
                futures["conda_env"].result()

    @timer
    def fast_setup_environment(architecture, storage, env, prefix, pkgs_dir):
        install_fast_initializer(architecture)

        # Get package urls
        conda_pkgs = env["conda"]
        pypi_pkgs = env.get("pypi", [])
        conda_pkg_urls = [package["path"] for package in conda_pkgs]
        pypi_pkg_urls = [package["path"] for package in pypi_pkgs]

        # Create string with package URLs
        all_package_urls = ""
        for url in conda_pkg_urls:
            all_package_urls += f"{storage.datastore_root}/{url}\n"
        all_package_urls += "---\n"
        for url in pypi_pkg_urls:
            all_package_urls += f"{storage.datastore_root}/{url}\n"

        # Initialize environment
        # NOTE: For the time being the fast-initializer only works for the S3 datastore implementation
        cmd = f"fast-initializer --prefix {prefix} --packages-dir {pkgs_dir}"
        run_cmd(cmd, all_package_urls)

    if len(sys.argv) != 4:
        print("Usage: bootstrap.py <flow_name> <id> <datastore_type>")
        sys.exit(1)

    try:
        _, flow_name, id_, datastore_type = sys.argv

        system = platform.system().lower()
        arch_machine = platform.machine().lower()

        if system == "darwin" and arch_machine == "arm64":
            architecture = "osx-arm64"
        elif system == "darwin":
            architecture = "osx-64"
        elif system == "linux" and arch_machine == "aarch64":
            architecture = "linux-aarch64"
        else:
            # default fallback
            architecture = "linux-64"

        prefix = os.path.join(os.getcwd(), architecture, id_)
        pkgs_dir = os.path.join(os.getcwd(), ".pkgs")
        conda_pkgs_dir = os.path.join(pkgs_dir, "conda")
        pypi_pkgs_dir = os.path.join(pkgs_dir, "pypi")
        manifest_dir = os.path.join(os.getcwd(), DATASTORE_LOCAL_DIR, flow_name)

        datastores = [d for d in DATASTORES if d.TYPE == datastore_type]
        if not datastores:
            print(f"No datastore found for type: {datastore_type}")
            sys.exit(1)

        storage = datastores[0](
            _datastore_packageroot(datastores[0], lambda *args, **kwargs: None)
        )

        # Move MAGIC_FILE inside local datastore.
        os.makedirs(manifest_dir, exist_ok=True)
        path_to_manifest = MetaflowCodeContent.get_filename(
            MAGIC_FILE, ContentType.OTHER_CONTENT
        )
        if path_to_manifest is None:
            raise RuntimeError(f"Cannot find {MAGIC_FILE} in the package")
        shutil.move(
            path_to_manifest,
            os.path.join(manifest_dir, MAGIC_FILE),
        )
        with open(os.path.join(manifest_dir, MAGIC_FILE)) as f:
            env = json.load(f)[id_][architecture]

        if CONDA_USE_FAST_INIT:
            fast_setup_environment(architecture, storage, env, prefix, pkgs_dir)
        else:
            setup_environment(
                architecture, storage, env, prefix, conda_pkgs_dir, pypi_pkgs_dir
            )

    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)



import json
import os
from metaflow.decorators import StepDecorator
from metaflow.metadata_provider import MetaDatum

from .airflow_utils import (
    TASK_ID_XCOM_KEY,
    FOREACH_CARDINALITY_XCOM_KEY,
)

K8S_XCOM_DIR_PATH = "/airflow/xcom"


def safe_mkdir(dir):
    try:
        os.makedirs(dir)
    except FileExistsError:
        pass


def push_xcom_values(xcom_dict):
    safe_mkdir(K8S_XCOM_DIR_PATH)
    with open(os.path.join(K8S_XCOM_DIR_PATH, "return.json"), "w") as f:
        json.dump(xcom_dict, f)


class AirflowInternalDecorator(StepDecorator):
    name = "airflow_internal"

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        meta = {}
        meta["airflow-dag-run-id"] = os.environ["METAFLOW_AIRFLOW_DAG_RUN_ID"]
        meta["airflow-job-id"] = os.environ["METAFLOW_AIRFLOW_JOB_ID"]
        entries = [
            MetaDatum(
                field=k, value=v, type=k, tags=["attempt_id:{0}".format(retry_count)]
            )
            for k, v in meta.items()
        ]

        # Register book-keeping metadata for debugging.
        metadata.register_metadata(run_id, step_name, task_id, entries)

    def task_finished(
        self, step_name, flow, graph, is_task_ok, retry_count, max_user_code_retries
    ):
        # This will pass the xcom when the task finishes.
        xcom_values = {
            TASK_ID_XCOM_KEY: os.environ["METAFLOW_AIRFLOW_TASK_ID"],
        }
        if graph[step_name].type == "foreach":
            xcom_values[FOREACH_CARDINALITY_XCOM_KEY] = flow._foreach_num_splits
        push_xcom_values(xcom_values)



import os
import ast
import time
import asyncio
import tempfile
import select
import fcntl
from contextlib import contextmanager
from subprocess import CalledProcessError
from typing import Any, Dict, TYPE_CHECKING, ContextManager, Tuple

if TYPE_CHECKING:
    import tempfile
    import metaflow.runner.subprocess_manager
    import metaflow.runner.click_api


def get_current_cell(ipython):
    if ipython:
        return ipython.history_manager.input_hist_raw[-1]
    return None


def format_flowfile(cell):
    """
    Formats the given cell content to create a valid Python script that can be
    executed as a Metaflow flow.
    """
    flowspec = [
        x
        for x in ast.parse(cell).body
        if isinstance(x, ast.ClassDef) and any(b.id == "FlowSpec" for b in x.bases)
    ]

    if not flowspec:
        raise ModuleNotFoundError(
            "The cell doesn't contain any class that inherits from 'FlowSpec'"
        )

    lines = cell.splitlines()[: flowspec[0].end_lineno]
    lines += ["if __name__ == '__main__':", f"    {flowspec[0].name}()"]
    return "\n".join(lines)


def check_process_exited(
    command_obj: "metaflow.runner.subprocess_manager.CommandManager",
) -> bool:
    if isinstance(command_obj.process, asyncio.subprocess.Process):
        return command_obj.process.returncode is not None
    else:
        return command_obj.process.poll() is not None


@contextmanager
def temporary_fifo() -> ContextManager[Tuple[str, int]]:
    """
    Create and open the read side of a temporary FIFO in a non-blocking mode.

    Returns
    -------
    str
        Path to the temporary FIFO.
    int
        File descriptor of the temporary FIFO.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "fifo")
        os.mkfifo(path)
        # Blocks until the write side is opened unless in non-blocking mode
        fd = os.open(path, os.O_RDONLY | os.O_NONBLOCK)
        try:
            yield path, fd
        finally:
            os.close(fd)


def read_from_fifo_when_ready(
    fifo_fd: int,
    command_obj: "metaflow.runner.subprocess_manager.CommandManager",
    encoding: str = "utf-8",
    timeout: int = 3600,
) -> str:
    """
    Read the content from the FIFO file descriptor when it is ready.

    Parameters
    ----------
    fifo_fd : int
        File descriptor of the FIFO.
    command_obj : CommandManager
        Command manager object that handles the write side of the FIFO.
    encoding : str, optional
        Encoding to use while reading the file, by default "utf-8".
    timeout : int, optional
        Timeout for reading the file in seconds, by default 3600.

    Returns
    -------
    str
        Content read from the FIFO.

    Raises
    ------
    TimeoutError
        If no event occurs on the FIFO within the timeout.
    CalledProcessError
        If the process managed by `command_obj` has exited without writing any
        content to the FIFO.
    """
    content = bytearray()
    poll = select.poll()
    poll.register(fifo_fd, select.POLLIN)
    while True:
        if check_process_exited(command_obj) and command_obj.process.returncode != 0:
            raise CalledProcessError(
                command_obj.process.returncode, command_obj.command
            )

        if timeout < 0:
            raise TimeoutError("Timeout while waiting for the file content")

        poll_begin = time.time()
        # We poll for a very short time to be also able to check if the file was closed
        # If the file is closed, we assume that we only have one writer so if we have
        # data, we break out. This is to work around issues in macos
        events = poll.poll(min(10, timeout * 1000))
        timeout -= time.time() - poll_begin

        try:
            data = os.read(fifo_fd, 8192)
            if data:
                content += data
                # We got data! Now switch to blocking mode for guaranteed complete reads.
                # In blocking mode, read() won't return 0 until writer closes AND all
                # kernel buffers are drained - this is POSIX guaranteed.
                flags = fcntl.fcntl(fifo_fd, fcntl.F_GETFL)
                fcntl.fcntl(fifo_fd, fcntl.F_SETFL, flags & ~os.O_NONBLOCK)

                # Now do blocking reads until true EOF
                while True:
                    chunk = os.read(fifo_fd, 8192)
                    if not chunk:
                        # True EOF - all data drained
                        break
                    content += chunk
                # All data read, exit main loop
                break
            else:
                if len(events):
                    # We read an EOF -- consider the file done
                    break
                else:
                    # We had no events (just a timeout) and the read didn't return
                    # an exception so the file is still open; we continue waiting for data
                    pass
        except BlockingIOError:
            # File not ready yet, continue waiting
            pass

    if not content and check_process_exited(command_obj):
        raise CalledProcessError(command_obj.process.returncode, command_obj.command)

    return content.decode(encoding)


async def async_read_from_fifo_when_ready(
    fifo_fd: int,
    command_obj: "metaflow.runner.subprocess_manager.CommandManager",
    encoding: str = "utf-8",
    timeout: int = 3600,
) -> str:
    """
    Read the content from the FIFO file descriptor when it is ready.

    Parameters
    ----------
    fifo_fd : int
        File descriptor of the FIFO.
    command_obj : CommandManager
        Command manager object that handles the write side of the FIFO.
    encoding : str, optional
        Encoding to use while reading the file, by default "utf-8".
    timeout : int, optional
        Timeout for reading the file in seconds, by default 3600.

    Returns
    -------
    str
        Content read from the FIFO.

    Raises
    ------
    TimeoutError
        If no event occurs on the FIFO within the timeout.
    CalledProcessError
        If the process managed by `command_obj` has exited without writing any
        content to the FIFO.
    """
    return await asyncio.to_thread(
        read_from_fifo_when_ready, fifo_fd, command_obj, encoding, timeout
    )


def make_process_error_message(
    command_obj: "metaflow.runner.subprocess_manager.CommandManager",
):
    stdout_log = open(command_obj.log_files["stdout"], encoding="utf-8").read()
    stderr_log = open(command_obj.log_files["stderr"], encoding="utf-8").read()
    command = " ".join(command_obj.command)
    error_message = "Error executing: '%s':\n" % command
    if stdout_log.strip():
        error_message += "\nStdout:\n%s\n" % stdout_log
    if stderr_log.strip():
        error_message += "\nStderr:\n%s\n" % stderr_log
    return error_message


def handle_timeout(
    attribute_file_fd: int,
    command_obj: "metaflow.runner.subprocess_manager.CommandManager",
    file_read_timeout: int,
):
    """
    Handle the timeout for a running subprocess command that reads a file
    and raises an error with appropriate logs if a TimeoutError occurs.

    Parameters
    ----------
    attribute_file_fd : int
        File descriptor belonging to the FIFO containing the attribute data.
    command_obj : CommandManager
        Command manager object that encapsulates the running command details.
    file_read_timeout : int
        Timeout for reading the file, in seconds

    Returns
    -------
    str
        Content read from the temporary file.

    Raises
    ------
    RuntimeError
        If a TimeoutError occurs, it raises a RuntimeError with the command's
        stdout and stderr logs.
    """
    try:
        return read_from_fifo_when_ready(
            attribute_file_fd, command_obj=command_obj, timeout=file_read_timeout
        )
    except (CalledProcessError, TimeoutError) as e:
        raise RuntimeError(make_process_error_message(command_obj)) from e


async def async_handle_timeout(
    attribute_file_fd: "int",
    command_obj: "metaflow.runner.subprocess_manager.CommandManager",
    file_read_timeout: int,
):
    """
    Handle the timeout for a running subprocess command that reads a file
    and raises an error with appropriate logs if a TimeoutError occurs.

    Parameters
    ----------
    attribute_file_fd : int
        File descriptor belonging to the FIFO containing the attribute data.
    command_obj : CommandManager
        Command manager object that encapsulates the running command details.
    file_read_timeout : int
        Timeout for reading the file, in seconds

    Returns
    -------
    str
        Content read from the temporary file.

    Raises
    ------
    RuntimeError
        If a TimeoutError occurs, it raises a RuntimeError with the command's
        stdout and stderr logs.
    """
    try:
        return await async_read_from_fifo_when_ready(
            attribute_file_fd, command_obj=command_obj, timeout=file_read_timeout
        )
    except (CalledProcessError, TimeoutError) as e:
        raise RuntimeError(make_process_error_message(command_obj)) from e


def get_lower_level_group(
    api: "metaflow.runner.click_api.MetaflowAPI",
    top_level_kwargs: Dict[str, Any],
    sub_command: str,
    sub_command_kwargs: Dict[str, Any],
) -> "metaflow.runner.click_api.MetaflowAPI":
    """
    Retrieve a lower-level group from the API based on the type and provided arguments.

    Parameters
    ----------
    api : MetaflowAPI
        Metaflow API instance.
    top_level_kwargs : Dict[str, Any]
        Top-level keyword arguments to pass to the API.
    sub_command : str
        Sub-command of API to get the API for
    sub_command_kwargs : Dict[str, Any]
        Sub-command arguments

    Returns
    -------
    MetaflowAPI
        The lower-level group object retrieved from the API.

    Raises
    ------
    ValueError
        If the `_type` is None.
    """
    sub_command_obj = getattr(api(**top_level_kwargs), sub_command)

    if sub_command_obj is None:
        raise ValueError(f"Sub-command '{sub_command}' not found in API '{api.name}'")

    return sub_command_obj(**sub_command_kwargs)


@contextmanager
def with_dir(new_dir):
    current_dir = os.getcwd()
    os.chdir(new_dir)
    yield new_dir
    os.chdir(current_dir)
